"""Model class for stable diffusion2."""
from typing import Optional

import torch
from diffusers import AutoPipelineForImage2Image, AutoPipelineForText2Image, StableDiffusionPipeline
from PIL.Image import Image

from .util import get_generator, clear_cache, get_logger

logger = get_logger(__name__)


class SD2Turbo:

    base_model: StableDiffusionPipeline

    def __init__(self,
                 base_model_id: str = "stabilityai/sd-turbo",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True,
                 guidance_scale: float = 0.0,
                 num_inference_steps: int = 1,
                 max_sequence_length: int = 256):
        self.base_model = AutoPipelineForText2Image.from_pretrained(
            base_model_id,
            use_safetensors=True,
            variant=variant,
            torch_dtype=torch_dtype,
            device_map=device_map,
            low_cpu_mem_usage=low_cpu_mem_usage
        )
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.max_sequence_length = max_sequence_length

    def __call__(self,
                 prompt: str,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: Optional[float] = None,
                 num_inference_steps: Optional[int] = None,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42,
                 max_sequence_length: Optional[int] = None) -> Image:
        guidance_scale = self.guidance_scale if guidance_scale is None else guidance_scale
        num_inference_steps = self.num_inference_steps if num_inference_steps is None else num_inference_steps
        max_sequence_length = self.max_sequence_length if max_sequence_length is None else max_sequence_length
        output_list = self.base_model(
            prompt=prompt,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            max_sequence_length=max_sequence_length,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)


class SD2TurboImg2Img:

    base_model: StableDiffusionPipeline

    def __init__(self,
                 base_model_id: str = "stabilityai/sd-turbo",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True,
                 guidance_scale: float = 0.0,
                 num_inference_steps: int = 2,
                 strength: float = 0.5,
                 max_sequence_length: int = 256):
        self.base_model = AutoPipelineForImage2Image.from_pretrained(
            base_model_id,
            use_safetensors=True,
            variant=variant,
            torch_dtype=torch_dtype,
            device_map=device_map,
            low_cpu_mem_usage=low_cpu_mem_usage
        )
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.strength = strength
        self.max_sequence_length = max_sequence_length

    def __call__(self,
                 prompt: str,
                 image: Image,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: Optional[float] = None,
                 num_inference_steps: Optional[int] = None,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42,
                 strength: Optional[float] = None,
                 max_sequence_length: Optional[int] = None) -> Image:
        guidance_scale = self.guidance_scale if guidance_scale is None else guidance_scale
        num_inference_steps = self.num_inference_steps if num_inference_steps is None else num_inference_steps
        strength = self.strength if strength is None else strength
        max_sequence_length = self.max_sequence_length if max_sequence_length is None else max_sequence_length
        output_list = self.base_model(
            image=image,
            prompt=prompt,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            strength=strength,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            max_sequence_length=max_sequence_length,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



"""Model class for controlnet.
The best image resolution 1024 x 1024.
"""
from typing import Optional, Callable
import torch
from diffusers import StableDiffusion3ControlNetPipeline, SD3ControlNetModel
from PIL.Image import Image
from .util import get_generator, clear_cache, get_logger


logger = get_logger(__name__)


class ControlNetSD3:

    base_model: StableDiffusion3ControlNetPipeline
    get_condition: Callable[[Image], Image]

    def __init__(self,
                 base_model_id: str = "stabilityai/stable-diffusion-3-medium-diffusers",
                 condition_type: str = "canny",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: Optional[str] = "balanced",
                 low_cpu_mem_usage: bool = True,
                 enable_model_cpu_offload: bool = False):
        config = dict(use_safetensors=True, torch_dtype=torch_dtype, low_cpu_mem_usage=low_cpu_mem_usage)
        if condition_type == "canny":
            controlnet = SD3ControlNetModel.from_pretrained("InstantX/SD3-Controlnet-Canny", **config)
        elif condition_type == "pose":
            controlnet = SD3ControlNetModel.from_pretrained("InstantX/SD3-Controlnet-Pose", **config)
        elif condition_type == "tile":
            controlnet = SD3ControlNetModel.from_pretrained("InstantX/SD3-Controlnet-Tile", **config)
        else:
            raise ValueError(f"unknown condition: {condition_type}")
        if torch.cuda.is_available():
            controlnet = controlnet.cuda()
        self.base_model = StableDiffusion3ControlNetPipeline.from_pretrained(
            base_model_id, controlnet=controlnet, variant=variant, device_map=device_map, **config
        )
        if enable_model_cpu_offload:
            self.base_model.enable_model_cpu_offload()

    def __call__(self,
                 prompt: str,
                 image: Image,
                 controlnet_conditioning_scale: float = 0.5,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: float = 7,
                 num_inference_steps: int = 28,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42) -> Image:
        output_list = self.base_model(
            prompt=prompt,
            control_image=image,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



"""Model class for stable depth2image."""
from typing import Optional
import torch
from diffusers import StableDiffusionDepth2ImgPipeline
from PIL.Image import Image
from .util import get_generator, clear_cache, get_logger
from .model_img2img_depth_anything_v2 import DepthAnythingV2

logger = get_logger(__name__)


class Depth2Image:

    base_model: StableDiffusionDepth2ImgPipeline

    def __init__(self,
                 base_model_id: str = "stabilityai/stable-diffusion-2-depth",
                 variant: Optional[str] = "fp16",
                 torch_dtype: Optional[torch.dtype] = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True):
        self.base_model = StableDiffusionDepth2ImgPipeline.from_pretrained(
            base_model_id, use_safetensors=True, variant=variant, torch_dtype=torch_dtype, device_map=device_map, low_cpu_mem_usage=low_cpu_mem_usage
        )
        self.depth_anything = DepthAnythingV2()

    def __call__(self,
                 image: Image,
                 prompt: str,
                 depth_maps: Optional[torch.Tensor] = None,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: float = 7.5,
                 num_inference_steps: int = 50,
                 num_images_per_prompt: int = 1,
                 seed: int = 42) -> Image:
        if depth_maps is None:
            logger.info("run depth anything v2")
            depth_maps = self.depth_anything(image, return_tensor=True)
        logger.info("run depth2image")
        output_list = self.base_model(
            prompt=prompt,
            image=image,
            depth_map=depth_maps.unsqueeze(0),
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_inference_steps=num_inference_steps,
            num_images_per_prompt=num_images_per_prompt,
            generator=get_generator(seed)
        ).images[0]
        clear_cache()
        return output_list

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



# image generation models
from .model_txt2img_image_stable_diffusion_xl import (
    SDXL,
    SDXLBase,
    SDXLBaseImg2Img,
    SDXLTurbo,
    SDXLTurboImg2Img,
    RealVisXL,
    RealVisXLImg2Img
)
from .model_txt2img_image_stable_diffusion_3 import (
    SD3,
    SD3Medium,
    SD3Large,
    SD3LargeTurbo,
    SD3BitsAndBytesModel,
    SD3LargeBitsAndBytesModel,
    SD3LargeTurboBitsAndBytesModel
)
from .model_txt2img_image_stable_diffusion_2 import SD2Turbo, SD2TurboImg2Img
from .model_txt2img_flux_1_dev import Flux1Dev

# super resolution/image restoration models
from .model_img2img_instruct_ir import InstructIR
from .model_img2img_stable_diffusion_upscaler import SDUpScaler

# image to text prompt interrogation
from .model_img2txt_clip_interrogator import CLIPInterrogator

# conditional image generation models
from .model_img2img_depth2image import Depth2Image
from .model_img2img_controlnet_stable_diffusion_2 import ControlNetSD2
from .model_img2img_controlnet_stable_diffusion_3 import ControlNetSD3
from .model_img2img_ledits_pp import LEditsPP

# depth
from .model_img2img_depth_anything_v2 import DepthAnythingV2



"""Model class for stable diffusion2."""
from typing import Optional, List, Union
import torch
from diffusers import DiffusionPipeline, StableDiffusionXLPipeline, StableDiffusionXLImg2ImgPipeline
from PIL.Image import Image
from .util import get_generator, clear_cache, get_logger, resize_image

logger = get_logger(__name__)


class SDXL:

    height: Optional[int]
    width: Optional[int]
    guidance_scale: float
    num_inference_steps: float
    high_noise_frac: float
    strength: float
    img2img: bool
    base_model_id: str
    base_model: Union[StableDiffusionXLPipeline, StableDiffusionXLImg2ImgPipeline]
    refiner_model: Optional[StableDiffusionXLPipeline]

    def __init__(self,
                 use_refiner: bool = True,
                 base_model_id: str = "stabilityai/stable-diffusion-xl-base-1.0",
                 refiner_model_id: str = "stabilityai/stable-diffusion-xl-refiner-1.0",
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 guidance_scale: float = 5.0,
                 num_inference_steps: int = 40,
                 high_noise_frac: float = 0.8,
                 strength: float = 0.5,
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 img2img: bool = False,
                 low_cpu_mem_usage: bool = True):
        config = dict(
            use_safetensors=True,
            variant=variant,
            torch_dtype=torch_dtype,
            device_map=device_map,
            low_cpu_mem_usage=low_cpu_mem_usage
        )
        self.height = height
        self.width = width
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.high_noise_frac = high_noise_frac
        self.strength = strength
        self.img2img = img2img
        if self.img2img:
            self.base_model = StableDiffusionXLImg2ImgPipeline.from_pretrained(base_model_id, **config)
        else:
            self.base_model = StableDiffusionXLPipeline.from_pretrained(base_model_id, **config)
        self.refiner_model = None
        if use_refiner:
            self.refiner_model = DiffusionPipeline.from_pretrained(refiner_model_id, text_encoder_2=self.base_model.text_encoder_2, vae=self.base_model.vae, **config)

    def validate_input(self,
                       prompt: Optional[List[str]] = None,
                       image: Optional[List[Image]] = None,
                       negative_prompt: Optional[List[str]] = None) -> None:
        if prompt is None:
            raise ValueError("No prompt provided.")
        if self.img2img:
            if image is None:
                raise ValueError("No image provided for img2img generation.")
            if len(image) != len(prompt):
                raise ValueError(f"Wrong shape: {len(image)} != {len(prompt)}")
        else:
            if image is not None:
                raise ValueError("Text2img cannot take image input.")
        if negative_prompt is not None:
            if len(prompt) != len(negative_prompt):
                raise ValueError(f"Wrong shape: {len(prompt)} != {len(negative_prompt)}")

    def __call__(self,
                 prompt: Optional[str] = None,
                 image: Optional[Image] = None,
                 strength: Optional[float] = None,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: Optional[float] = None,
                 num_inference_steps: Optional[int] = None,
                 num_images_per_prompt: int = 1,
                 high_noise_frac: Optional[float] = None,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: Optional[int] = None) -> Image:
        self.validate_input(prompt, image, negative_prompt)
        shared_config = dict(
            num_inference_steps=self.num_inference_steps if num_inference_steps is None else num_inference_steps,
            num_images_per_prompt=num_images_per_prompt,
            height=height if height is not None else self.height,
            width=width if width is not None else self.width,
            generator=get_generator(seed),
            guidance_scale=self.guidance_scale if guidance_scale is None else guidance_scale,
        )
        if self.img2img:
            shared_config["strength"] = self.strength if strength is None else strength
            if shared_config["height"] is not None and shared_config["width"] is not None:
                image = resize_image(image, shared_config["width"], shared_config["height"])
        if self.refiner_model:
            shared_config["output_type"] = "latent"
            shared_config["denoising_end"] = self.high_noise_frac if high_noise_frac is None else high_noise_frac
        if self.img2img:
            output = self.base_model(
                image=image,
                prompt=prompt,
                negative_prompt=negative_prompt,
                **shared_config
            ).images
        else:
            output = self.base_model(
                prompt=prompt,
                negative_prompt=negative_prompt,
                **shared_config
            ).images
        if self.refiner_model:
            output_list = self.refiner_model(
                image=output,
                prompt=prompt,
                negative_prompt=negative_prompt,
                num_inference_steps=shared_config["num_inference_steps"],
                denoising_start=shared_config["denoising_end"],
                height=shared_config["height"],
                width=shared_config["width"],
                generator=shared_config["generator"]
            ).images
        else:
            output_list = output
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)


class SDXLTurbo(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=False,
            base_model_id="stabilityai/sdxl-turbo",
            guidance_scale=0.0,
            num_inference_steps=1,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True
        )


class SDXLTurboImg2Img(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=False,
            base_model_id="stabilityai/sdxl-turbo",
            height=512,
            width=512,
            guidance_scale=0.0,
            num_inference_steps=2,
            strength=0.5,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True,
            img2img=True,
        )


class SDXLBase(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=True,
            base_model_id="stabilityai/stable-diffusion-xl-base-1.0",
            refiner_model_id="stabilityai/stable-diffusion-xl-refiner-1.0",
            guidance_scale=5.0,
            num_inference_steps=40,
            high_noise_frac=0.8,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True
        )


class SDXLBaseImg2Img(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=True,
            base_model_id="stabilityai/stable-diffusion-xl-base-1.0",
            refiner_model_id="stabilityai/stable-diffusion-xl-refiner-1.0",
            guidance_scale=5.0,
            num_inference_steps=80,
            high_noise_frac=0.8,
            strength=0.5,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True,
            img2img=True
        )


class RealVisXL(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=False,
            base_model_id="SG161222/RealVisXL_V4.0",
            guidance_scale=3.0,
            num_inference_steps=20,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True,
        )


class RealVisXLImg2Img(SDXL):

    def __init__(self):
        super().__init__(
            use_refiner=False,
            base_model_id="SG161222/RealVisXL_V4.0",
            guidance_scale=3.0,
            num_inference_steps=20,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True,
            strength=0.5,
            img2img=True
        )


"""Model class for stable diffusion3."""
from typing import Optional

import torch
from diffusers import StableDiffusion3Pipeline
from PIL.Image import Image
from diffusers import BitsAndBytesConfig, SD3Transformer2DModel
from transformers import T5EncoderModel

from .util import get_generator, clear_cache, get_logger

logger = get_logger(__name__)


class SD3:

    base_model: StableDiffusion3Pipeline

    def __init__(self,
                 base_model_id: str = "stabilityai/stable-diffusion-3-medium-diffusers",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True,
                 guidance_scale: float = 7.0,
                 num_inference_steps: int = 28,
                 max_sequence_length: int = 256):
        self.base_model = StableDiffusion3Pipeline.from_pretrained(
            base_model_id,
            use_safetensors=True,
            variant=variant,
            torch_dtype=torch_dtype,
            device_map=device_map,
            low_cpu_mem_usage=low_cpu_mem_usage
        )
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.max_sequence_length = max_sequence_length

    def __call__(self,
                 prompt: str,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: Optional[float] = None,
                 num_inference_steps: Optional[int] = None,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42,
                 max_sequence_length: Optional[int] = None) -> Image:
        guidance_scale = self.guidance_scale if guidance_scale is None else guidance_scale
        num_inference_steps = self.num_inference_steps if num_inference_steps is None else num_inference_steps
        max_sequence_length = self.max_sequence_length if max_sequence_length is None else max_sequence_length
        output_list = self.base_model(
            prompt=prompt,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            max_sequence_length=max_sequence_length,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)


class SD3Medium(SD3):

    def __init__(self):
        super().__init__(
            "stabilityai/stable-diffusion-3-medium-diffusers",
            guidance_scale=7.0,
            num_inference_steps=28,
            variant="fp16",
            torch_dtype=torch.float16,
            device_map="balanced",
            low_cpu_mem_usage=True,
            max_sequence_length=256
        )


class SD3Large(SD3):

    def __init__(self):
        super().__init__(
            "stabilityai/stable-diffusion-3.5-large",
            guidance_scale=3.5,
            num_inference_steps=28,
            variant="fp16",
            torch_dtype=torch.bfloat16,
            device_map="balanced",
            low_cpu_mem_usage=True,
            max_sequence_length=256
        )


class SD3LargeTurbo(SD3):

    def __init__(self):
        super().__init__(
            "stabilityai/stable-diffusion-3.5-large-turbo",
            guidance_scale=0.0,
            num_inference_steps=4,
            variant="fp16",
            torch_dtype=torch.bfloat16,
            device_map="balanced",
            low_cpu_mem_usage=True
        )


class SD3BitsAndBytesModel:

    def __init__(self,
                 model_id: str = "stabilityai/stable-diffusion-3.5-large-turbo",
                 num_inference_steps: int = 4,
                 guidance_scale: float = 0.0,
                 max_sequence_length: int = 256):
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.max_sequence_length = max_sequence_length
        self.nf4_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16
        )
        self.model_nf4 = SD3Transformer2DModel.from_pretrained(
            model_id,
            subfolder="transformer",
            quantization_config=self.nf4_config,
            torch_dtype=torch.bfloat16
        )
        self.t5_nf4 = T5EncoderModel.from_pretrained(
            "diffusers/t5-nf4", torch_dtype=torch.bfloat16
        )
        self.base_model = StableDiffusion3Pipeline.from_pretrained(
            model_id,
            transformer=self.model_nf4,
            text_encoder_3=self.t5_nf4,
            torch_dtype=torch.bfloat16
        )
        self.base_model.enable_model_cpu_offload()

    def __call__(self,
                 prompt: str,
                 negative_prompt: Optional[str] = None,
                 num_inference_steps: Optional[int] = None,
                 guidance_scale: Optional[float] = None,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42,
                 max_sequence_length: Optional[int] = None) -> Image:
        guidance_scale = self.guidance_scale if guidance_scale is None else guidance_scale
        num_inference_steps = self.num_inference_steps if num_inference_steps is None else num_inference_steps
        max_sequence_length = self.max_sequence_length if max_sequence_length is None else max_sequence_length
        output_list = self.base_model(
            prompt=prompt,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            max_sequence_length=max_sequence_length,
            width=width,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)


class SD3LargeBitsAndBytesModel(SD3BitsAndBytesModel):

    def __init__(self):
        super().__init__(
            model_id="stabilityai/stable-diffusion-3.5-large",
            num_inference_steps=28,
            guidance_scale=4.5,
            max_sequence_length=256
        )


class SD3LargeTurboBitsAndBytesModel(SD3BitsAndBytesModel):

    def __init__(self):
        super().__init__(
            model_id="stabilityai/stable-diffusion-3.5-large-turbo",
            num_inference_steps=4,
            guidance_scale=0.0,
            max_sequence_length=256
        )



"""Model class for stable diffusion upscaler (note that this model suits for 128 x 128). """
from typing import Optional, List
import torch
from diffusers import StableDiffusionUpscalePipeline
from PIL.Image import Image
from .util import clear_cache, get_logger, resize_image

logger = get_logger(__name__)


class SDUpScaler:

    base_model: StableDiffusionUpscalePipeline
    height: int = 128
    width: int = 128

    def __init__(self,
                 base_model_id: str = "stabilityai/stable-diffusion-x4-upscaler",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True):
        self.base_model = StableDiffusionUpscalePipeline.from_pretrained(
            base_model_id, use_safetensors=True, variant=variant, torch_dtype=torch_dtype, device_map=device_map, low_cpu_mem_usage=low_cpu_mem_usage
        )

    def __call__(self,
                 image: Image,
                 prompt: Optional[str] = None,
                 reshape_method: Optional[str] = "downscale",
                 upscale_factor: int = 4) -> List[Image]:


        def downscale_image(x: Image) -> Image:
            return resize_image(x, width=int(x.width / upscale_factor), height=int(x.height / upscale_factor))

        if reshape_method == "best":
            image = resize_image(image, width=self.width, height=self.height)
        elif reshape_method == "downscale":
            image = downscale_image(image)
        elif reshape_method is not None:
            raise ValueError(f"unknown reshape method: {reshape_method}")
        output_list = self.base_model(image=image, prompt=prompt).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



"""Model class for stable DepthAnythingV2."""
from typing import Union
import torch
import numpy as np
from diffusers import StableDiffusion3Pipeline
from transformers import pipeline
from PIL.Image import Image, fromarray
from .util import clear_cache, get_logger

logger = get_logger(__name__)


def tensor_to_image(predicted_depth: torch.Tensor, image: Image) -> Image:
    size = image.size[::-1]
    depth = predicted_depth.unsqueeze(0).unsqueeze(0)
    prediction = torch.nn.functional.interpolate(depth, size=size, mode="bicubic", align_corners=False)
    prediction = (prediction - prediction.min()) / (prediction.max() - prediction.min()) * 255.0
    prediction = prediction.numpy()[0][0].astype(np.uint8)
    return fromarray(prediction)


def reverse_value(predicted_depth: torch.Tensor) -> torch.Tensor:
    return (predicted_depth - predicted_depth.max()) * -1


class DepthAnythingV2:

    base_model: StableDiffusion3Pipeline

    def __init__(self,
                 base_model_id: str = "depth-anything/Depth-Anything-V2-Large-hf",
                 torch_dtype: torch.dtype = torch.float32):
        if torch.cuda.is_available():
            self.pipe = pipeline(task="depth-estimation", model=base_model_id, torch_dtype=torch_dtype, device="cuda")
        else:
            self.pipe = pipeline(task="depth-estimation", model=base_model_id)

    def __call__(self,
                 image: Image,
                 return_tensor: bool = False,
                 reverse_depth: bool = False) -> Union[Image, torch.Tensor]:
        depth = self.pipe(image)
        clear_cache()
        if reverse_depth:
            depth = reverse_value(depth["predicted_depth"])
        else:
            depth = depth["predicted_depth"]
        if return_tensor:
            return depth
        return tensor_to_image(depth, image)

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



"""Model class for FLUX 1 dev."""
from typing import Optional
import torch
from diffusers import FluxPipeline
from PIL.Image import Image
from .util import get_generator, clear_cache, get_logger

logger = get_logger(__name__)


class Flux1Dev:

    base_model: FluxPipeline

    def __init__(self,
                 base_model_id: str = "black-forest-labs/FLUX.1-dev",
                 torch_dtype: torch.dtype = torch.bfloat16,
                 device_map: Optional[str] = "balanced",
                 low_cpu_mem_usage: bool = True,
                 enable_model_cpu_offload: bool = False):
        self.base_model = FluxPipeline.from_pretrained(
            base_model_id, use_safetensors=True, torch_dtype=torch_dtype, device_map=device_map, low_cpu_mem_usage=low_cpu_mem_usage
        )
        if enable_model_cpu_offload:
            self.base_model.enable_model_cpu_offload()

    def __call__(self,
                 prompt: str,
                 guidance_scale: float = 3.5,
                 num_inference_steps: int = 28,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42) -> Image:
        output_list = self.base_model(
            prompt=prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)



"""Model class for controlnet."""
from typing import Optional, Callable
import torch
import numpy as np
from cv2 import Canny
from diffusers import StableDiffusionXLControlNetPipeline, ControlNetModel, AutoencoderKL
from PIL.Image import Image, fromarray
from .util import get_generator, clear_cache, get_logger
from .model_img2img_depth_anything_v2 import DepthAnythingV2

logger = get_logger(__name__)


def get_canny_edge(image: Image) -> Image:
    image = Canny(np.array(image), 100, 200)
    image = image[:, :, None]
    image = np.concatenate([image, image, image], axis=2)
    return fromarray(image)


class ControlNetSD2:

    base_model: ControlNetModel
    get_condition: Callable[[Image], Image]

    def __init__(self,
                 base_model_id: str = "stabilityai/stable-diffusion-xl-base-1.0",
                 vae_model_id: str = "madebyollin/sdxl-vae-fp16-fix",
                 condition_type: str = "canny",
                 variant: str = "fp16",
                 torch_dtype: torch.dtype = torch.float16,
                 device_map: str = "balanced",
                 low_cpu_mem_usage: bool = True):
        vae = AutoencoderKL.from_pretrained(
            vae_model_id, torch_dtype=torch_dtype, device_map=device_map, low_cpu_mem_usage=low_cpu_mem_usage
        )
        if condition_type == "canny":
            controlnet = ControlNetModel.from_pretrained(
                "diffusers/controlnet-canny-sdxl-1.0", use_safetensors=True, torch_dtype=torch_dtype, variant=variant, low_cpu_mem_usage=low_cpu_mem_usage
            )
            self.get_condition = get_canny_edge
        elif condition_type == "depth":
            controlnet = ControlNetModel.from_pretrained(
                "diffusers/controlnet-zoe-depth-sdxl-1.0", torch_dtype=torch_dtype, low_cpu_mem_usage=low_cpu_mem_usage
            )
            depth = DepthAnythingV2()
            self.get_condition = lambda x: depth([x])[0]
        else:
            raise ValueError(f"unknown condition: {condition_type}")
        if torch.cuda.is_available():
            controlnet = controlnet.cuda()
        self.base_model = StableDiffusionXLControlNetPipeline.from_pretrained(
            base_model_id, controlnet=controlnet, vae=vae, use_safetensors=True, device_map=device_map, torch_dtype=torch_dtype, variant=variant, low_cpu_mem_usage=low_cpu_mem_usage
        )

    def __call__(self,
                 prompt: str,
                 image: Image,
                 controlnet_conditioning_scale: float = 0.5,
                 negative_prompt: Optional[str] = None,
                 guidance_scale: float = 7.5,
                 num_inference_steps: int = 40,
                 num_images_per_prompt: int = 1,
                 height: Optional[int] = None,
                 width: Optional[int] = None,
                 seed: int = 42) -> Image:
        logger.info("generate condition")
        condition = self.get_condition(image)
        logger.info("generate image")
        output_list = self.base_model(
            prompt=prompt,
            image=condition,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            num_inference_steps=num_inference_steps,
            height=height,
            width=width,
            generator=get_generator(seed)
        ).images
        clear_cache()
        return output_list[0]

    @staticmethod
    def export(data: Image, output_path: str, file_format: str = "png") -> None:
        data.save(output_path, file_format)
